@@ -22,8 +22,8 @@
 #include "gtkeditable.h"
 #include "gtkentry.h"
 #include "gtkmarshalers.h"
-#include "gtksignal.h"
 #include "gtkintl.h"
+#include "gtktreeprivate.h"
 
 static void gtk_cell_renderer_text_init       (GtkCellRendererText      *celltext);
 static void gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class);
@@ -71,6 +71,7 @@ enum {
   PROP_TEXT,
   PROP_MARKUP,
   PROP_ATTRIBUTES,
+  PROP_SINGLE_PARAGRAPH_MODE,
   
   /* Style args */
   PROP_BACKGROUND,
@@ -91,6 +92,7 @@ enum {
   PROP_STRIKETHROUGH,
   PROP_UNDERLINE,
   PROP_RISE,
+  PROP_LANGUAGE,
   
   /* Whether-a-style-arg-is-set args */
   PROP_BACKGROUND_SET,
@@ -105,7 +107,8 @@ enum {
   PROP_EDITABLE_SET,
   PROP_STRIKETHROUGH_SET,
   PROP_UNDERLINE_SET,
-  PROP_RISE_SET
+  PROP_RISE_SET,
+  PROP_LANGUAGE_SET
 };
 
 static gpointer parent_class;
@@ -113,10 +116,24 @@ static guint text_cell_renderer_signals [LAST_SIGNAL];
 
 #define GTK_CELL_RENDERER_TEXT_PATH "gtk-cell-renderer-text-path"
 
-GtkType
+#define GTK_CELL_RENDERER_TEXT_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GTK_TYPE_CELL_RENDERER_TEXT, GtkCellRendererTextPrivate))
+
+typedef struct _GtkCellRendererTextPrivate GtkCellRendererTextPrivate;
+struct _GtkCellRendererTextPrivate
+{
+  guint single_paragraph : 1;
+  guint language_set : 1;
+  guint markup_set : 1;
+
+  gulong focus_out_id;
+  PangoLanguage *language;
+};
+
+
+GType
 gtk_cell_renderer_text_get_type (void)
 {
-  static GtkType cell_text_type = 0;
+  static GType cell_text_type = 0;
 
   if (!cell_text_type)
     {
@@ -133,7 +150,9 @@ gtk_cell_renderer_text_get_type (void)
         (GInstanceInitFunc) gtk_cell_renderer_text_init,
       };
 
-      cell_text_type = g_type_register_static (GTK_TYPE_CELL_RENDERER, "GtkCellRendererText", &cell_text_info, 0);
+      cell_text_type =
+	g_type_register_static (GTK_TYPE_CELL_RENDERER, "GtkCellRendererText",
+				&cell_text_info, 0);
     }
 
   return cell_text_type;
@@ -170,56 +189,65 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_TEXT,
                                    g_param_spec_string ("text",
-                                                        _("Text"),
-                                                        _("Text to render"),
+                                                        P_("Text"),
+                                                        P_("Text to render"),
                                                         NULL,
                                                         G_PARAM_READWRITE));
   
   g_object_class_install_property (object_class,
                                    PROP_MARKUP,
                                    g_param_spec_string ("markup",
-                                                        _("Markup"),
-                                                        _("Marked up text to render"),
+                                                        P_("Markup"),
+                                                        P_("Marked up text to render"),
                                                         NULL,
                                                         G_PARAM_WRITABLE));
 
   g_object_class_install_property (object_class,
 				   PROP_ATTRIBUTES,
 				   g_param_spec_boxed ("attributes",
-						       _("Attributes"),
-						       _("A list of style attributes to apply to the text of the renderer."),
+						       P_("Attributes"),
+						       P_("A list of style attributes to apply to the text of the renderer"),
 						       PANGO_TYPE_ATTR_LIST,
 						       G_PARAM_READWRITE));
+
+  g_object_class_install_property (object_class,
+                                   PROP_SINGLE_PARAGRAPH_MODE,
+                                   g_param_spec_boolean ("single_paragraph_mode",
+                                                         P_("Single Paragraph Mode"),
+                                                         P_("Whether or not to keep all text in a single paragraph"),
+                                                         FALSE,
+                                                         G_PARAM_READWRITE));
+
   
   g_object_class_install_property (object_class,
                                    PROP_BACKGROUND,
                                    g_param_spec_string ("background",
-                                                        _("Background color name"),
-                                                        _("Background color as a string"),
+                                                        P_("Background color name"),
+                                                        P_("Background color as a string"),
                                                         NULL,
                                                         G_PARAM_WRITABLE));
 
   g_object_class_install_property (object_class,
                                    PROP_BACKGROUND_GDK,
                                    g_param_spec_boxed ("background_gdk",
-                                                       _("Background color"),
-                                                       _("Background color as a GdkColor"),
+                                                       P_("Background color"),
+                                                       P_("Background color as a GdkColor"),
                                                        GDK_TYPE_COLOR,
                                                        G_PARAM_READABLE | G_PARAM_WRITABLE));  
 
   g_object_class_install_property (object_class,
                                    PROP_FOREGROUND,
                                    g_param_spec_string ("foreground",
-                                                        _("Foreground color name"),
-                                                        _("Foreground color as a string"),
+                                                        P_("Foreground color name"),
+                                                        P_("Foreground color as a string"),
                                                         NULL,
                                                         G_PARAM_WRITABLE));
 
   g_object_class_install_property (object_class,
                                    PROP_FOREGROUND_GDK,
                                    g_param_spec_boxed ("foreground_gdk",
-                                                       _("Foreground color"),
-                                                       _("Foreground color as a GdkColor"),
+                                                       P_("Foreground color"),
+                                                       P_("Foreground color as a GdkColor"),
                                                        GDK_TYPE_COLOR,
                                                        G_PARAM_READABLE | G_PARAM_WRITABLE));
 
@@ -227,24 +255,24 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_EDITABLE,
                                    g_param_spec_boolean ("editable",
-                                                         _("Editable"),
-                                                         _("Whether the text can be modified by the user"),
+                                                         P_("Editable"),
+                                                         P_("Whether the text can be modified by the user"),
                                                          FALSE,
                                                          G_PARAM_READABLE | G_PARAM_WRITABLE));
 
   g_object_class_install_property (object_class,
                                    PROP_FONT,
                                    g_param_spec_string ("font",
-                                                        _("Font"),
-                                                        _("Font description as a string"),
+                                                        P_("Font"),
+                                                        P_("Font description as a string"),
                                                         NULL,
                                                         G_PARAM_READABLE | G_PARAM_WRITABLE));
 
   g_object_class_install_property (object_class,
                                    PROP_FONT_DESC,
                                    g_param_spec_boxed ("font_desc",
-                                                       _("Font"),
-                                                       _("Font description as a PangoFontDescription struct"),
+                                                       P_("Font"),
+                                                       P_("Font description as a PangoFontDescription struct"),
                                                        PANGO_TYPE_FONT_DESCRIPTION,
                                                        G_PARAM_READABLE | G_PARAM_WRITABLE));
 
@@ -252,16 +280,16 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_FAMILY,
                                    g_param_spec_string ("family",
-                                                        _("Font family"),
-                                                        _("Name of the font family, e.g. Sans, Helvetica, Times, Monospace"),
+                                                        P_("Font family"),
+                                                        P_("Name of the font family, e.g. Sans, Helvetica, Times, Monospace"),
                                                         NULL,
                                                         G_PARAM_READABLE | G_PARAM_WRITABLE));
 
   g_object_class_install_property (object_class,
                                    PROP_STYLE,
                                    g_param_spec_enum ("style",
-                                                      _("Font style"),
-                                                      _("Font style"),
+                                                      P_("Font style"),
+                                                      P_("Font style"),
                                                       PANGO_TYPE_STYLE,
                                                       PANGO_STYLE_NORMAL,
                                                       G_PARAM_READABLE | G_PARAM_WRITABLE));
@@ -269,8 +297,8 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_VARIANT,
                                    g_param_spec_enum ("variant",
-                                                     _("Font variant"),
-                                                     _("Font variant"),
+                                                     P_("Font variant"),
+                                                     P_("Font variant"),
                                                       PANGO_TYPE_VARIANT,
                                                       PANGO_VARIANT_NORMAL,
                                                       G_PARAM_READABLE | G_PARAM_WRITABLE));
@@ -278,8 +306,8 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_WEIGHT,
                                    g_param_spec_int ("weight",
-                                                     _("Font weight"),
-                                                     _("Font weight"),
+                                                     P_("Font weight"),
+                                                     P_("Font weight"),
                                                      0,
                                                      G_MAXINT,
                                                      PANGO_WEIGHT_NORMAL,
@@ -288,8 +316,8 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
    g_object_class_install_property (object_class,
                                    PROP_STRETCH,
                                    g_param_spec_enum ("stretch",
-                                                      _("Font stretch"),
-                                                      _("Font stretch"),
+                                                      P_("Font stretch"),
+                                                      P_("Font stretch"),
                                                       PANGO_TYPE_STRETCH,
                                                       PANGO_STRETCH_NORMAL,
                                                       G_PARAM_READABLE | G_PARAM_WRITABLE));
@@ -297,8 +325,8 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_SIZE,
                                    g_param_spec_int ("size",
-                                                     _("Font size"),
-                                                     _("Font size"),
+                                                     P_("Font size"),
+                                                     P_("Font size"),
                                                      0,
                                                      G_MAXINT,
                                                      0,
@@ -307,8 +335,8 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_SIZE_POINTS,
                                    g_param_spec_double ("size_points",
-                                                        _("Font points"),
-                                                        _("Font size in points"),
+                                                        P_("Font points"),
+                                                        P_("Font size in points"),
                                                         0.0,
                                                         G_MAXDOUBLE,
                                                         0.0,
@@ -317,8 +345,8 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_SCALE,
                                    g_param_spec_double ("scale",
-                                                        _("Font scale"),
-                                                        _("Font scaling factor"),
+                                                        P_("Font scale"),
+                                                        P_("Font scaling factor"),
                                                         0.0,
                                                         G_MAXDOUBLE,
                                                         1.0,
@@ -327,8 +355,8 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_RISE,
                                    g_param_spec_int ("rise",
-                                                     _("Rise"),
-                                                     _("Offset of text above the baseline (below the baseline if rise is negative)"),
+                                                     P_("Rise"),
+                                                     P_("Offset of text above the baseline (below the baseline if rise is negative)"),
                                                      -G_MAXINT,
                                                      G_MAXINT,
                                                      0,
@@ -338,92 +366,110 @@ gtk_cell_renderer_text_class_init (GtkCellRendererTextClass *class)
   g_object_class_install_property (object_class,
                                    PROP_STRIKETHROUGH,
                                    g_param_spec_boolean ("strikethrough",
-                                                         _("Strikethrough"),
-                                                         _("Whether to strike through the text"),
+                                                         P_("Strikethrough"),
+                                                         P_("Whether to strike through the text"),
                                                          FALSE,
                                                          G_PARAM_READABLE | G_PARAM_WRITABLE));
   
   g_object_class_install_property (object_class,
                                    PROP_UNDERLINE,
                                    g_param_spec_enum ("underline",
-                                                      _("Underline"),
-                                                      _("Style of underline for this text"),
+                                                      P_("Underline"),
+                                                      P_("Style of underline for this text"),
                                                       PANGO_TYPE_UNDERLINE,
                                                       PANGO_UNDERLINE_NONE,
                                                       G_PARAM_READABLE | G_PARAM_WRITABLE));
 
+  g_object_class_install_property (object_class,
+                                   PROP_LANGUAGE,
+                                   g_param_spec_string ("language",
+                                                        P_("Language"),
+                                                        P_("The language this text is in, as an ISO code. Pango can use this as a hint when rendering the text. If you don't understand this parameter, you probably don't need it"),
+                                                        NULL,
+                                                        G_PARAM_READWRITE));
+
+
   /* Style props are set or not */
 
 #define ADD_SET_PROP(propname, propval, nick, blurb) g_object_class_install_property (object_class, propval, g_param_spec_boolean (propname, nick, blurb, FALSE, G_PARAM_READABLE | G_PARAM_WRITABLE))
 
   ADD_SET_PROP ("background_set", PROP_BACKGROUND_SET,
-                _("Background set"),
-                _("Whether this tag affects the background color"));
+                P_("Background set"),
+                P_("Whether this tag affects the background color"));
 
   ADD_SET_PROP ("foreground_set", PROP_FOREGROUND_SET,
-                _("Foreground set"),
-                _("Whether this tag affects the foreground color"));
+                P_("Foreground set"),
+                P_("Whether this tag affects the foreground color"));
   
   ADD_SET_PROP ("editable_set", PROP_EDITABLE_SET,
-                _("Editability set"),
-                _("Whether this tag affects text editability"));
+                P_("Editability set"),
+                P_("Whether this tag affects text editability"));
 
   ADD_SET_PROP ("family_set", PROP_FAMILY_SET,
-                _("Font family set"),
-                _("Whether this tag affects the font family"));  
+                P_("Font family set"),
+                P_("Whether this tag affects the font family"));  
 
   ADD_SET_PROP ("style_set", PROP_STYLE_SET,
-                _("Font style set"),
-                _("Whether this tag affects the font style"));
+                P_("Font style set"),
+                P_("Whether this tag affects the font style"));
 
   ADD_SET_PROP ("variant_set", PROP_VARIANT_SET,
-                _("Font variant set"),
-                _("Whether this tag affects the font variant"));
+                P_("Font variant set"),
+                P_("Whether this tag affects the font variant"));
 
   ADD_SET_PROP ("weight_set", PROP_WEIGHT_SET,
-                _("Font weight set"),
-                _("Whether this tag affects the font weight"));
+                P_("Font weight set"),
+                P_("Whether this tag affects the font weight"));
 
   ADD_SET_PROP ("stretch_set", PROP_STRETCH_SET,
-                _("Font stretch set"),
-                _("Whether this tag affects the font stretch"));
+                P_("Font stretch set"),
+                P_("Whether this tag affects the font stretch"));
 
   ADD_SET_PROP ("size_set", PROP_SIZE_SET,
-                _("Font size set"),
-                _("Whether this tag affects the font size"));
+                P_("Font size set"),
+                P_("Whether this tag affects the font size"));
 
   ADD_SET_PROP ("scale_set", PROP_SCALE_SET,
-                _("Font scale set"),
-                _("Whether this tag scales the font size by a factor"));
+                P_("Font scale set"),
+                P_("Whether this tag scales the font size by a factor"));
   
   ADD_SET_PROP ("rise_set", PROP_RISE_SET,
-                _("Rise set"),
-                _("Whether this tag affects the rise"));
+                P_("Rise set"),
+                P_("Whether this tag affects the rise"));
 
   ADD_SET_PROP ("strikethrough_set", PROP_STRIKETHROUGH_SET,
-                _("Strikethrough set"),
-                _("Whether this tag affects strikethrough"));
+                P_("Strikethrough set"),
+                P_("Whether this tag affects strikethrough"));
 
   ADD_SET_PROP ("underline_set", PROP_UNDERLINE_SET,
-                _("Underline set"),
-                _("Whether this tag affects underlining"));
+                P_("Underline set"),
+                P_("Whether this tag affects underlining"));
+
+  ADD_SET_PROP ("language_set", PROP_LANGUAGE_SET,
+                P_("Language set"),
+                P_("Whether this tag affects the language the text is rendered as"));
 
   text_cell_renderer_signals [EDITED] =
-    gtk_signal_new ("edited",
-		    GTK_RUN_LAST,
-		    GTK_CLASS_TYPE (object_class),
-		    GTK_SIGNAL_OFFSET (GtkCellRendererTextClass, edited),
-		    _gtk_marshal_VOID__STRING_STRING,
-		    GTK_TYPE_NONE, 2,
-		    G_TYPE_STRING,
-		    G_TYPE_STRING);
+    g_signal_new ("edited",
+		  G_OBJECT_CLASS_TYPE (object_class),
+		  G_SIGNAL_RUN_LAST,
+		  G_STRUCT_OFFSET (GtkCellRendererTextClass, edited),
+		  NULL, NULL,
+		  _gtk_marshal_VOID__STRING_STRING,
+		  G_TYPE_NONE, 2,
+		  G_TYPE_STRING,
+		  G_TYPE_STRING);
 
+  g_type_class_add_private (object_class, sizeof (GtkCellRendererTextPrivate));
 }
 
 static void
 gtk_cell_renderer_text_finalize (GObject *object)
 {
   GtkCellRendererText *celltext = GTK_CELL_RENDERER_TEXT (object);
+  GtkCellRendererTextPrivate *priv;
+
+  priv = GTK_CELL_RENDERER_TEXT_GET_PRIVATE (object);
 
   pango_font_description_free (celltext->font);
 
@@ -433,6 +479,9 @@ gtk_cell_renderer_text_finalize (GObject *object)
   if (celltext->extra_attrs)
     pango_attr_list_unref (celltext->extra_attrs);
 
+  if (priv->language)
+    g_object_unref (priv->language);
+
   (* G_OBJECT_CLASS (parent_class)->finalize) (object);
 }
 
@@ -465,6 +514,9 @@ gtk_cell_renderer_text_get_property (GObject        *object,
 				     GParamSpec     *pspec)
 {
   GtkCellRendererText *celltext = GTK_CELL_RENDERER_TEXT (object);
+  GtkCellRendererTextPrivate *priv;
+
+  priv = GTK_CELL_RENDERER_TEXT_GET_PRIVATE (object);
 
   switch (param_id)
     {
@@ -476,6 +528,10 @@ gtk_cell_renderer_text_get_property (GObject        *object,
       g_value_set_boxed (value, celltext->extra_attrs);
       break;
 
+    case PROP_SINGLE_PARAGRAPH_MODE:
+      g_value_set_boolean (value, priv->single_paragraph);
+      break;
+
     case PROP_BACKGROUND_GDK:
       {
         GdkColor color;
@@ -563,6 +619,10 @@ gtk_cell_renderer_text_get_property (GObject        *object,
       g_value_set_int (value, celltext->rise);
       break;  
 
+    case PROP_LANGUAGE:
+      g_value_set_string (value, pango_language_to_string (priv->language));
+      break;
+
     case PROP_BACKGROUND_SET:
       g_value_set_boolean (value, celltext->background_set);
       break;
@@ -603,6 +663,10 @@ gtk_cell_renderer_text_get_property (GObject        *object,
     case  PROP_RISE_SET:
       g_value_set_boolean (value, celltext->rise_set);
       break;
+
+    case PROP_LANGUAGE_SET:
+      g_value_set_boolean (value, priv->language_set);
+      break;
       
     case PROP_BACKGROUND:
     case PROP_FOREGROUND:
@@ -798,12 +862,24 @@ gtk_cell_renderer_text_set_property (GObject      *object,
 				     GParamSpec   *pspec)
 {
   GtkCellRendererText *celltext = GTK_CELL_RENDERER_TEXT (object);
+  GtkCellRendererTextPrivate *priv;
+
+  priv = GTK_CELL_RENDERER_TEXT_GET_PRIVATE (object);
 
   switch (param_id)
     {
     case PROP_TEXT:
       if (celltext->text)
         g_free (celltext->text);
+
+      if (priv->markup_set)
+        {
+          if (celltext->extra_attrs)
+            pango_attr_list_unref (celltext->extra_attrs);
+          celltext->extra_attrs = NULL;
+          priv->markup_set = FALSE;
+        }
+
       celltext->text = g_strdup (g_value_get_string (value));
       g_object_notify (object, "text");
       break;
@@ -822,12 +898,6 @@ gtk_cell_renderer_text_set_property (GObject      *object,
 	gchar *text = NULL;
 	GError *error = NULL;
 	PangoAttrList *attrs = NULL;
-	
-	if (celltext->text)
-	  g_free (celltext->text);
-
-	if (celltext->extra_attrs)
-	  pango_attr_list_unref (celltext->extra_attrs);
 
 	str = g_value_get_string (value);
 	if (str && !pango_parse_markup (str,
@@ -843,11 +913,22 @@ gtk_cell_renderer_text_set_property (GObject      *object,
 	    g_error_free (error);
 	    return;
 	  }
-	
+
+	if (celltext->text)
+	  g_free (celltext->text);
+
+	if (celltext->extra_attrs)
+	  pango_attr_list_unref (celltext->extra_attrs);
+
 	celltext->text = text;
 	celltext->extra_attrs = attrs;
+        priv->markup_set = TRUE;
       }
       break;
+
+    case PROP_SINGLE_PARAGRAPH_MODE:
+      priv->single_paragraph = g_value_get_boolean (value);
+      break;
       
     case PROP_BACKGROUND:
       {
@@ -1006,6 +1087,14 @@ gtk_cell_renderer_text_set_property (GObject      *object,
 	celltext->calc_fixed_height = TRUE;
       break;  
 
+    case PROP_LANGUAGE:
+      priv->language_set = TRUE;
+      if (priv->language)
+        g_object_unref (priv->language);
+      priv->language = pango_language_from_string (g_value_get_string (value));
+      g_object_notify (object, "language_set");
+      break;
+
     case PROP_BACKGROUND_SET:
       celltext->background_set = g_value_get_boolean (value);
       break;
@@ -1055,6 +1144,10 @@ gtk_cell_renderer_text_set_property (GObject      *object,
       celltext->rise_set = g_value_get_boolean (value);
       break;
 
+    case PROP_LANGUAGE_SET:
+      priv->language_set = g_value_get_boolean (value);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
       break;
@@ -1077,7 +1170,7 @@ gtk_cell_renderer_text_set_property (GObject      *object,
 GtkCellRenderer *
 gtk_cell_renderer_text_new (void)
 {
-  return GTK_CELL_RENDERER (g_object_new (gtk_cell_renderer_text_get_type (), NULL));
+  return g_object_new (GTK_TYPE_CELL_RENDERER_TEXT, NULL);
 }
 
 static void
@@ -1099,6 +1192,9 @@ get_layout (GtkCellRendererText *celltext,
   PangoAttrList *attr_list;
   PangoLayout *layout;
   PangoUnderline uline;
+  GtkCellRendererTextPrivate *priv;
+
+  priv = GTK_CELL_RENDERER_TEXT_GET_PRIVATE (celltext);
   
   layout = gtk_widget_create_pango_layout (widget, celltext->text);
 
@@ -1107,6 +1203,8 @@ get_layout (GtkCellRendererText *celltext,
   else
     attr_list = pango_attr_list_new ();
 
+  pango_layout_set_single_paragraph_mode (layout, priv->single_paragraph);
+
   if (will_render)
     {
       /* Add options that affect appearance but not size */
@@ -1140,6 +1238,9 @@ get_layout (GtkCellRendererText *celltext,
     uline = celltext->underline_style;
   else
     uline = PANGO_UNDERLINE_NONE;
+
+  if (priv->language_set)
+    add_attr (attr_list, pango_attr_language_new (priv->language));
   
   if ((flags & GTK_CELL_RENDERER_PRELIT) == GTK_CELL_RENDERER_PRELIT)
     {
@@ -1207,7 +1308,9 @@ gtk_cell_renderer_text_get_size (GtkCellRenderer *cell,
       row_height = (pango_font_metrics_get_ascent (metrics) +
 		    pango_font_metrics_get_descent (metrics));
       pango_font_metrics_unref (metrics);
-      
+
+      pango_font_description_free (font_desc);
+
       gtk_cell_renderer_set_fixed_size (cell,
 					cell->width, 2*cell->ypad +
 					celltext->fixed_height_rows * PANGO_PIXELS (row_height));
@@ -1234,7 +1337,8 @@ gtk_cell_renderer_text_get_size (GtkCellRenderer *cell,
     {
       if (x_offset)
 	{
-	  *x_offset = cell->xalign * (cell_area->width - rect.width - (2 * cell->xpad));
+	  *x_offset = ((gtk_widget_get_direction (widget) == GTK_TEXT_DIR_RTL) ?
+		       (1.0 - cell->xalign) : cell->xalign) * (cell_area->width - rect.width - (2 * cell->xpad));
 	  *x_offset = MAX (*x_offset, 0);
 	}
       if (y_offset)
@@ -1244,12 +1348,12 @@ gtk_cell_renderer_text_get_size (GtkCellRenderer *cell,
 	}
     }
 
-  g_object_unref (G_OBJECT (layout));
+  g_object_unref (layout);
 }
 
 static void
 gtk_cell_renderer_text_render (GtkCellRenderer      *cell,
-			       GdkWindow            *window,
+			       GdkDrawable          *window,
 			       GtkWidget            *widget,
 			       GdkRectangle         *background_area,
 			       GdkRectangle         *cell_area,
@@ -1267,6 +1371,7 @@ gtk_cell_renderer_text_render (GtkCellRenderer      *cell,
 
   gtk_cell_renderer_text_get_size (cell, widget, cell_area, &x_offset, &y_offset, NULL, NULL);
 
+
   if ((flags & GTK_CELL_RENDERER_SELECTED) == GTK_CELL_RENDERER_SELECTED)
     {
       if (GTK_WIDGET_HAS_FOCUS (widget))
@@ -1294,7 +1399,9 @@ gtk_cell_renderer_text_render (GtkCellRenderer      *cell,
       gc = gdk_gc_new (window);
 
       gdk_gc_set_rgb_fg_color (gc, &color);
-      
+
+      if (expose_area)               
+        gdk_gc_set_clip_rectangle (gc, expose_area);
       gdk_draw_rectangle (window,
                           gc,
                           TRUE,
@@ -1302,37 +1409,62 @@ gtk_cell_renderer_text_render (GtkCellRenderer      *cell,
                           background_area->y,
                           background_area->width,
                           background_area->height);
-
-      g_object_unref (G_OBJECT (gc));
+      if (expose_area)               
+        gdk_gc_set_clip_rectangle (gc, NULL);
+      g_object_unref (gc);
     }
 
   gtk_paint_layout (widget->style,
                     window,
                     state,
 		    TRUE,
-                    cell_area,
+                    expose_area,
                     widget,
                     "cellrenderertext",
                     cell_area->x + x_offset + cell->xpad,
                     cell_area->y + y_offset + cell->ypad,
                     layout);
 
-  g_object_unref (G_OBJECT (layout));
+  g_object_unref (layout);
 }
+
 static void
 gtk_cell_renderer_text_editing_done (GtkCellEditable *entry,
 				     gpointer         data)
 {
   const gchar *path;
   const gchar *new_text;
+  GtkCellRendererTextPrivate *priv;
+
+  priv = GTK_CELL_RENDERER_TEXT_GET_PRIVATE (data);
+
+  if (priv->focus_out_id > 0)
+    {
+      g_signal_handler_disconnect (entry, priv->focus_out_id);
+      priv->focus_out_id = 0;
+    }
 
   if (GTK_ENTRY (entry)->editing_canceled)
-    return;
+    {
+      gtk_cell_renderer_editing_canceled (GTK_CELL_RENDERER (data));
+      return;
+    }
 
   path = g_object_get_data (G_OBJECT (entry), GTK_CELL_RENDERER_TEXT_PATH);
   new_text = gtk_entry_get_text (GTK_ENTRY (entry));
 
-  gtk_signal_emit (GTK_OBJECT (data), text_cell_renderer_signals[EDITED], path, new_text);
+  g_signal_emit (data, text_cell_renderer_signals[EDITED], 0, path, new_text);
+}
+
+static gboolean
+gtk_cell_renderer_text_focus_out_event (GtkWidget *entry,
+		                        GdkEvent  *event,
+					gpointer   data)
+{
+  gtk_cell_renderer_text_editing_done (GTK_CELL_EDITABLE (entry), data);
+
+  /* entry needs focus-out-event */
+  return FALSE;
 }
 
 static GtkCellEditable *
@@ -1344,10 +1476,12 @@ gtk_cell_renderer_text_start_editing (GtkCellRenderer      *cell,
 				      GdkRectangle         *cell_area,
 				      GtkCellRendererState  flags)
 {
-  GtkCellRendererText *celltext;
   GtkWidget *entry;
-  
+  GtkCellRendererText *celltext;
+  GtkCellRendererTextPrivate *priv;
+
   celltext = GTK_CELL_RENDERER_TEXT (cell);
+  priv = GTK_CELL_RENDERER_TEXT_GET_PRIVATE (cell);
 
   /* If the cell isn't editable we return NULL. */
   if (celltext->editable == FALSE)
@@ -1364,10 +1498,14 @@ gtk_cell_renderer_text_start_editing (GtkCellRenderer      *cell,
   gtk_editable_select_region (GTK_EDITABLE (entry), 0, -1);
   
   gtk_widget_show (entry);
-  gtk_signal_connect (GTK_OBJECT (entry),
-		      "editing_done",
-		      G_CALLBACK (gtk_cell_renderer_text_editing_done),
-		      celltext);
+  g_signal_connect (entry,
+		    "editing_done",
+		    G_CALLBACK (gtk_cell_renderer_text_editing_done),
+		    celltext);
+  priv->focus_out_id = g_signal_connect (entry, "focus_out_event",
+		         G_CALLBACK (gtk_cell_renderer_text_focus_out_event),
+		         celltext);
+
   return GTK_CELL_EDITABLE (entry);
 
 }
