@@ -35,6 +35,12 @@
 #include "gtkintl.h"
 #include "gtkbindings.h"
 
+#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GTK_TYPE_DIALOG, GtkDialogPrivate))
+
+typedef struct {
+  guint ignore_separator : 1;
+} GtkDialogPrivate;
+
 typedef struct _ResponseData ResponseData;
 
 struct _ResponseData
@@ -130,11 +136,13 @@ gtk_dialog_class_init (GtkDialogClass *class)
 
   class->close = gtk_dialog_close;
   
+  g_type_class_add_private (gobject_class, sizeof (GtkDialogPrivate));
+
   g_object_class_install_property (gobject_class,
                                    PROP_HAS_SEPARATOR,
                                    g_param_spec_boolean ("has_separator",
-							 _("Has separator"),
-							 _("The dialog has a separator bar above its buttons"),
+							 P_("Has separator"),
+							 P_("The dialog has a separator bar above its buttons"),
                                                          TRUE,
                                                          G_PARAM_READWRITE));
   
@@ -159,16 +167,16 @@ gtk_dialog_class_init (GtkDialogClass *class)
   
   gtk_widget_class_install_style_property (widget_class,
 					   g_param_spec_int ("content_area_border",
-                                                             _("Content area border"),
-                                                             _("Width of border around the main dialog area"),
+                                                             P_("Content area border"),
+                                                             P_("Width of border around the main dialog area"),
                                                              0,
                                                              G_MAXINT,
                                                              2,
                                                              G_PARAM_READABLE));
   gtk_widget_class_install_style_property (widget_class,
                                            g_param_spec_int ("button_spacing",
-                                                             _("Button spacing"),
-                                                             _("Spacing between buttons"),
+                                                             P_("Button spacing"),
+                                                             P_("Spacing between buttons"),
                                                              0,
                                                              G_MAXINT,
                                                              10,
@@ -176,8 +184,8 @@ gtk_dialog_class_init (GtkDialogClass *class)
   
   gtk_widget_class_install_style_property (widget_class,
                                            g_param_spec_int ("action_area_border",
-                                                             _("Action area border"),
-                                                             _("Width of border around the button area at the bottom of the dialog"),
+                                                             P_("Action area border"),
+                                                             P_("Width of border around the button area at the bottom of the dialog"),
                                                              0,
                                                              G_MAXINT,
                                                              5,
@@ -219,6 +227,11 @@ update_spacings (GtkDialog *dialog)
 static void
 gtk_dialog_init (GtkDialog *dialog)
 {
+  GtkDialogPrivate *priv;
+
+  priv = GET_PRIVATE (dialog);
+  priv->ignore_separator = FALSE;
+
   /* To avoid breaking old code that prevents destroy on delete event
    * by connecting a handler, we have to have the FIRST signal
    * connection on the dialog.
@@ -766,10 +779,20 @@ void
 gtk_dialog_set_has_separator (GtkDialog *dialog,
                               gboolean   setting)
 {
+  GtkDialogPrivate *priv;
+
   g_return_if_fail (GTK_IS_DIALOG (dialog));
 
+  priv = GET_PRIVATE (dialog);
+
   /* this might fail if we get called before _init() somehow */
   g_assert (dialog->vbox != NULL);
+
+  if (priv->ignore_separator)
+    {
+      g_warning ("Ignoring the separator setting");
+      return;
+    }
   
   if (setting && dialog->separator == NULL)
     {
@@ -1001,3 +1024,13 @@ gtk_dialog_run (GtkDialog *dialog)
 
   return ri.response_id;
 }
+
+void
+_gtk_dialog_set_ignore_separator (GtkDialog *dialog,
+				  gboolean   ignore_separator)
+{
+  GtkDialogPrivate *priv;
+
+  priv = GET_PRIVATE (dialog);
+  priv->ignore_separator = ignore_separator;
+}
