@@ -32,6 +32,7 @@
 #include <gtk/gtkeditable.h>
 #include <gtk/gtkimcontext.h>
 #include <gtk/gtkmenu.h>
+#include <gtk/gtkentrycompletion.h>
 #include <pango/pango.h>
 
 #ifdef __cplusplus
@@ -40,11 +41,11 @@ extern "C" {
 
 
 #define GTK_TYPE_ENTRY                  (gtk_entry_get_type ())
-#define GTK_ENTRY(obj)                  (GTK_CHECK_CAST ((obj), GTK_TYPE_ENTRY, GtkEntry))
-#define GTK_ENTRY_CLASS(klass)          (GTK_CHECK_CLASS_CAST ((klass), GTK_TYPE_ENTRY, GtkEntryClass))
-#define GTK_IS_ENTRY(obj)               (GTK_CHECK_TYPE ((obj), GTK_TYPE_ENTRY))
-#define GTK_IS_ENTRY_CLASS(klass)       (GTK_CHECK_CLASS_TYPE ((klass), GTK_TYPE_ENTRY))
-#define GTK_ENTRY_GET_CLASS(obj)        (GTK_CHECK_GET_CLASS ((obj), GTK_TYPE_ENTRY, GtkEntryClass))
+#define GTK_ENTRY(obj)                  (G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_TYPE_ENTRY, GtkEntry))
+#define GTK_ENTRY_CLASS(klass)          (G_TYPE_CHECK_CLASS_CAST ((klass), GTK_TYPE_ENTRY, GtkEntryClass))
+#define GTK_IS_ENTRY(obj)               (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GTK_TYPE_ENTRY))
+#define GTK_IS_ENTRY_CLASS(klass)       (G_TYPE_CHECK_CLASS_TYPE ((klass), GTK_TYPE_ENTRY))
+#define GTK_ENTRY_GET_CLASS(obj)        (G_TYPE_INSTANCE_GET_CLASS ((obj), GTK_TYPE_ENTRY, GtkEntryClass))
 
 
 typedef struct _GtkEntry       GtkEntry;
@@ -90,6 +91,9 @@ struct _GtkEntry
 
   guint        mouse_cursor_obscured : 1;
   
+  guint        select_words : 1;
+  guint        select_lines : 1;
+  guint        resolved_dir : 4; /* PangoDirection */
   guint   button;
   guint   blink_timeout;
   guint   recompute_idle;
@@ -111,8 +115,6 @@ struct _GtkEntry
   gunichar invisible_char;
 
   gint width_chars;
-
-  GdkGC *cursor_gc;
 };
 
 struct _GtkEntryClass
@@ -147,7 +149,7 @@ struct _GtkEntryClass
   void (*_gtk_reserved4) (void);
 };
 
-GtkType    gtk_entry_get_type       		(void) G_GNUC_CONST;
+GType      gtk_entry_get_type       		(void) G_GNUC_CONST;
 GtkWidget* gtk_entry_new            		(void);
 void       gtk_entry_set_visibility 		(GtkEntry      *entry,
 						 gboolean       visible);
@@ -181,6 +183,13 @@ PangoLayout* gtk_entry_get_layout               (GtkEntry      *entry);
 void         gtk_entry_get_layout_offsets       (GtkEntry      *entry,
                                                  gint          *x,
                                                  gint          *y);
+void       gtk_entry_set_alignment              (GtkEntry      *entry,
+                                                 gfloat         xalign);
+gfloat     gtk_entry_get_alignment              (GtkEntry      *entry);
+
+void                gtk_entry_set_completion (GtkEntry           *entry,
+                                              GtkEntryCompletion *completion);
+GtkEntryCompletion *gtk_entry_get_completion (GtkEntry           *entry);
 
 /* Deprecated compatibility functions
  */
