@@ -106,12 +106,6 @@ gtk_file_chooser_class_init (gpointer g_iface)
 							    P_("The current filter for selecting which files are displayed"),
 							    GTK_TYPE_FILE_FILTER,
 							    G_PARAM_READWRITE));
-  g_object_interface_install_property (g_iface,
-				       g_param_spec_boolean ("folder-mode",
-							     P_("Folder Mode"),
-							     P_("Whether to select folders rather than files"),
-							     FALSE,
-							     G_PARAM_READWRITE));
   g_object_interface_install_property (g_iface,
 				       g_param_spec_boolean ("local-only",
 							     P_("Local Only"),
@@ -130,6 +124,12 @@ gtk_file_chooser_class_init (gpointer g_iface)
 							     P_("Whether the application supplied widget for custom previews should be shown."),
 							     TRUE,
 							     G_PARAM_READWRITE));
+  g_object_interface_install_property (g_iface,
+				       g_param_spec_boolean ("use-preview-label",
+							     P_("Use Preview Label"),
+							     P_("Whether to display a stock label with the name of the previewed file."),
+							     TRUE,
+							     G_PARAM_READWRITE));
   g_object_interface_install_property (g_iface,
 				       g_param_spec_object ("extra-widget",
 							    P_("Extra widget"),
@@ -214,53 +214,6 @@ gtk_file_chooser_get_action (GtkFileChooser *chooser)
   return action;
 }
 
-/**
- * gtk_file_chooser_set_folder_mode:
- * @chooser: a #GtkFileChooser
- * @folder_mode: %TRUE if the file chooser is used to select folders
- *               rather than files.
- * 
- * Sets whether the file chooser is used to select folders
- * rather than files. If in folder mode, only folders are displayed
- * to the use, and not the individual files inside the folders
- * and the user selects a single folder rather than one or
- * more files.
- *
- * Since: 2.4
- **/
-void
-gtk_file_chooser_set_folder_mode (GtkFileChooser *chooser,
-				  gboolean        folder_mode)
-{
-  g_return_if_fail (GTK_IS_FILE_CHOOSER (chooser));
-
-  g_object_set (chooser, "folder-mode", folder_mode, NULL);
-}
-
-/**
- * gtk_file_chooser_get_folder_mode:
- * @chooser: a #GtkFileChooser
- * 
- * Gets whether the file chooser is used to select folders
- * rather than files. See gtk_file_chooser_set_folder_mode()
- * 
- * Return value: %TRUE if the file chooser is used to select
- *               folders rather than files.
- *
- * Since: 2.4
- **/
-gboolean
-gtk_file_chooser_get_folder_mode (GtkFileChooser *chooser)
-{
-  gboolean folder_mode;
-  
-  g_return_val_if_fail (GTK_IS_FILE_CHOOSER (chooser), FALSE);
-
-  g_object_get (chooser, "folder-mode", &folder_mode, NULL);
-
-  return folder_mode;
-}
-
 /**
  * gtk_file_chooser_set_local_only:
  * @chooser: a #GtkFileChooser
@@ -315,10 +268,10 @@ gtk_file_chooser_get_local_only (GtkFileChooser *chooser)
  * @chooser: a #GtkFileChooser
  * @select_multiple: %TRUE if multiple files can be selected.
  * 
- * Sets whether multiple files can be selected in the file
- * selector. If the file selector if in folder mode (see
- * gtk_file_selector_set_folder_mode()) then only one folder
- * can be selected, without regard to this setting.
+ * Sets whether multiple files can be selected in the file selector.  This is
+ * only relevant if the action is set to be GTK_FILE_CHOOSER_ACTION_OPEN or
+ * GTK_FILE_CHOOSER_ACTION_SAVE.  It cannot be set with either of the folder
+ * actions.
  *
  * Since: 2.4
  **/
@@ -1100,12 +1053,11 @@ gtk_file_chooser_set_preview_widget_active (GtkFileChooser *chooser,
  * gtk_file_chooser_get_preview_widget_active:
  * @chooser: a #GtkFileChooser
  * 
- * Gets whether the preview widget set by
- * gtk_file_chooser_set_preview_widget_active() should be shown for the
- * current filename. See gtk_file_chooser_set_preview_widget_active().
+ * Gets whether the preview widget set by gtk_file_chooser_set_preview_widget()
+ * should be shown for the current filename. See
+ * gtk_file_chooser_set_preview_widget_active().
  * 
- * Return value: %TRUE if the preview widget is active for the
- *  current filename.
+ * Return value: %TRUE if the preview widget is active for the current filename.
  *
  * Since: 2.4
  **/
@@ -1121,6 +1073,51 @@ gtk_file_chooser_get_preview_widget_active (GtkFileChooser *chooser)
   return active;
 }
 
+/**
+ * gtk_file_chooser_set_use_preview_label:
+ * @chooser: a #GtkFileChooser
+ * @use_label: whether to display a stock label with the name of the previewed file
+ * 
+ * Sets whether the file chooser should display a stock label with the name of
+ * the file that is being previewed; the default is %TRUE.  Applications that
+ * want to draw the whole preview area themselves should set this to %FALSE and
+ * display the name themselves in their preview widget.
+ *
+ * See also: gtk_file_chooser_set_preview_widget()
+ *
+ * Since: 2.4
+ **/
+void
+gtk_file_chooser_set_use_preview_label (GtkFileChooser *chooser,
+					gboolean        use_label)
+{
+  g_return_if_fail (GTK_IS_FILE_CHOOSER (chooser));
+
+  g_object_set (chooser, "use-preview-label", use_label, NULL);
+}
+
+/**
+ * gtk_file_chooser_get_use_preview_label:
+ * @chooser: a #GtkFileChooser
+ * 
+ * Gets whether a stock label should be drawn with the name of the previewed
+ * file.  See gtk_file_chooser_set_use_preview_label().
+ * 
+ * Return value: %TRUE if the file chooser is set to display a label with the
+ * name of the previewed file, %FALSE otherwise.
+ **/
+gboolean
+gtk_file_chooser_get_use_preview_label (GtkFileChooser *chooser)
+{
+  gboolean use_label;
+  
+  g_return_val_if_fail (GTK_IS_FILE_CHOOSER (chooser), FALSE);
+
+  g_object_get (chooser, "use-preview-label", &use_label, NULL);
+
+  return use_label;
+}
+
 /**
  * gtk_file_chooser_get_preview_filename:
  * @chooser: a #GtkFileChooser
