@@ -37,7 +37,6 @@
 #include "gtkspinbutton.h"
 #include "gtkmain.h"
 #include "gtkmarshalers.h"
-#include "gtksignal.h"
 #include "gtksettings.h"
 #include "gtkintl.h"
 
@@ -146,23 +145,24 @@ static guint spinbutton_signals[LAST_SIGNAL] = {0};
 
 #define NO_ARROW 2
 
-GtkType
+GType
 gtk_spin_button_get_type (void)
 {
-  static GtkType spin_button_type = 0;
+  static GType spin_button_type = 0;
 
   if (!spin_button_type)
     {
-      static const GtkTypeInfo spin_button_info =
+      static const GTypeInfo spin_button_info =
       {
-	"GtkSpinButton",
-	sizeof (GtkSpinButton),
 	sizeof (GtkSpinButtonClass),
-	(GtkClassInitFunc) gtk_spin_button_class_init,
-	(GtkObjectInitFunc) gtk_spin_button_init,
-	/* reserved_1 */ NULL,
-        /* reserved_2 */ NULL,
-        (GtkClassInitFunc) NULL,
+	NULL,		/* base_init */
+	NULL,		/* base_finalize */
+	(GClassInitFunc) gtk_spin_button_class_init,
+	NULL,		/* class_finalize */
+	NULL,		/* class_data */
+	sizeof (GtkSpinButton),
+	0,		/* n_preallocs */
+	(GInstanceInitFunc) gtk_spin_button_init,
       };
 
       static const GInterfaceInfo editable_info =
@@ -172,7 +172,10 @@ gtk_spin_button_get_type (void)
 	NULL  /* interface_data */
       };
 
-      spin_button_type = gtk_type_unique (GTK_TYPE_ENTRY, &spin_button_info);
+      spin_button_type =
+	g_type_register_static (GTK_TYPE_ENTRY, "GtkSpinButton",
+				&spin_button_info, 0);
+
       g_type_add_interface_static (spin_button_type,
 				   GTK_TYPE_EDITABLE,
 				   &editable_info);
@@ -198,7 +201,7 @@ gtk_spin_button_class_init (GtkSpinButtonClass *class)
   widget_class   = (GtkWidgetClass*)   class;
   entry_class    = (GtkEntryClass*)    class;
 
-  parent_class = gtk_type_class (GTK_TYPE_ENTRY);
+  parent_class = g_type_class_peek_parent (class);
 
   gobject_class->finalize = gtk_spin_button_finalize;
 
@@ -234,16 +237,16 @@ gtk_spin_button_class_init (GtkSpinButtonClass *class)
   g_object_class_install_property (gobject_class,
                                    PROP_ADJUSTMENT,
                                    g_param_spec_object ("adjustment",
-                                                        _("Adjustment"),
-                                                        _("The adjustment that holds the value of the spinbutton"),
+                                                        P_("Adjustment"),
+                                                        P_("The adjustment that holds the value of the spinbutton"),
                                                         GTK_TYPE_ADJUSTMENT,
                                                         G_PARAM_READWRITE));
   
   g_object_class_install_property (gobject_class,
                                    PROP_CLIMB_RATE,
                                    g_param_spec_double ("climb_rate",
-							_("Climb Rate"),
-							_("The acceleration rate when you hold down a button"),
+							P_("Climb Rate"),
+							P_("The acceleration rate when you hold down a button"),
 							0.0,
 							G_MAXDOUBLE,
 							0.0,
@@ -252,8 +255,8 @@ gtk_spin_button_class_init (GtkSpinButtonClass *class)
   g_object_class_install_property (gobject_class,
                                    PROP_DIGITS,
                                    g_param_spec_uint ("digits",
-						      _("Digits"),
-						      _("The number of decimal places to display"),
+						      P_("Digits"),
+						      P_("The number of decimal places to display"),
 						      0,
 						      MAX_DIGITS,
 						      0,
@@ -262,32 +265,32 @@ gtk_spin_button_class_init (GtkSpinButtonClass *class)
   g_object_class_install_property (gobject_class,
                                    PROP_SNAP_TO_TICKS,
                                    g_param_spec_boolean ("snap_to_ticks",
-							 _("Snap to Ticks"),
-							 _("Whether erroneous values are automatically changed to a spin button's nearest step increment"),
+							 P_("Snap to Ticks"),
+							 P_("Whether erroneous values are automatically changed to a spin button's nearest step increment"),
 							 FALSE,
 							 G_PARAM_READWRITE));
   
   g_object_class_install_property (gobject_class,
                                    PROP_NUMERIC,
                                    g_param_spec_boolean ("numeric",
-							 _("Numeric"),
-							 _("Whether non-numeric characters should be ignored"),
+							 P_("Numeric"),
+							 P_("Whether non-numeric characters should be ignored"),
 							 FALSE,
 							 G_PARAM_READWRITE));
   
   g_object_class_install_property (gobject_class,
                                    PROP_WRAP,
                                    g_param_spec_boolean ("wrap",
-							 _("Wrap"),
-							 _("Whether a spin button should wrap upon reaching its limits"),
+							 P_("Wrap"),
+							 P_("Whether a spin button should wrap upon reaching its limits"),
 							 FALSE,
 							 G_PARAM_READWRITE));
   
   g_object_class_install_property (gobject_class,
                                    PROP_UPDATE_POLICY,
                                    g_param_spec_enum ("update_policy",
-						      _("Update Policy"),
-						      _("Whether the spin button should update always, or only when the value is legal"),
+						      P_("Update Policy"),
+						      P_("Whether the spin button should update always, or only when the value is legal"),
 						      GTK_TYPE_SPIN_BUTTON_UPDATE_POLICY,
 						      GTK_UPDATE_ALWAYS,
 						      G_PARAM_READWRITE));
@@ -295,48 +298,53 @@ gtk_spin_button_class_init (GtkSpinButtonClass *class)
   g_object_class_install_property (gobject_class,
                                    PROP_VALUE,
                                    g_param_spec_double ("value",
-							_("Value"),
-							_("Reads the current value, or sets a new value"),
+							P_("Value"),
+							P_("Reads the current value, or sets a new value"),
 							-G_MAXDOUBLE,
 							G_MAXDOUBLE,
 							0.0,
 							G_PARAM_READWRITE));  
   
   gtk_widget_class_install_style_property_parser (widget_class,
-						  g_param_spec_enum ("shadow_type", "Shadow Type", NULL,
+						  g_param_spec_enum ("shadow_type", 
+								     "Shadow Type", 
+								     P_("Style of bevel around the spin button"),
 								     GTK_TYPE_SHADOW_TYPE,
 								     GTK_SHADOW_IN,
 								     G_PARAM_READABLE),
 						  gtk_rc_property_parse_enum);
   spinbutton_signals[INPUT] =
-    gtk_signal_new ("input",
-		    GTK_RUN_LAST,
-		    GTK_CLASS_TYPE (object_class),
-		    GTK_SIGNAL_OFFSET (GtkSpinButtonClass, input),
-		    _gtk_marshal_INT__POINTER,
-		    GTK_TYPE_INT, 1, GTK_TYPE_POINTER);
+    g_signal_new ("input",
+		  G_TYPE_FROM_CLASS (gobject_class),
+		  G_SIGNAL_RUN_LAST,
+		  G_STRUCT_OFFSET (GtkSpinButtonClass, input),
+		  NULL, NULL,
+		  _gtk_marshal_INT__POINTER,
+		  G_TYPE_INT, 1,
+		  G_TYPE_POINTER);
 
   spinbutton_signals[OUTPUT] =
     g_signal_new ("output",
-                  G_TYPE_FROM_CLASS(object_class),
-                  G_SIGNAL_RUN_LAST,
-                  G_STRUCT_OFFSET(GtkSpinButtonClass, output),
-                  _gtk_boolean_handled_accumulator, NULL,
-                  _gtk_marshal_BOOLEAN__VOID,
-                  G_TYPE_BOOLEAN, 0);
+		  G_TYPE_FROM_CLASS (gobject_class),
+		  G_SIGNAL_RUN_LAST,
+		  G_STRUCT_OFFSET (GtkSpinButtonClass, output),
+		  _gtk_boolean_handled_accumulator, NULL,
+		  _gtk_marshal_BOOLEAN__VOID,
+		  G_TYPE_BOOLEAN, 0);
 
   spinbutton_signals[VALUE_CHANGED] =
-    gtk_signal_new ("value_changed",
-		    GTK_RUN_LAST,
-		    GTK_CLASS_TYPE (object_class),
-		    GTK_SIGNAL_OFFSET (GtkSpinButtonClass, value_changed),
-		    _gtk_marshal_VOID__VOID,
-		    GTK_TYPE_NONE, 0);
+    g_signal_new ("value_changed",
+		  G_TYPE_FROM_CLASS (gobject_class),
+		  G_SIGNAL_RUN_LAST,
+		  G_STRUCT_OFFSET (GtkSpinButtonClass, value_changed),
+		  NULL, NULL,
+		  _gtk_marshal_VOID__VOID,
+		  G_TYPE_NONE, 0);
 
   /* Action signals */
   spinbutton_signals[CHANGE_VALUE] =
     g_signal_new ("change_value",
-                  G_TYPE_FROM_CLASS (object_class),
+                  G_TYPE_FROM_CLASS (gobject_class),
                   G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
                   G_STRUCT_OFFSET (GtkSpinButtonClass, change_value),
                   NULL, NULL,
@@ -482,7 +490,7 @@ gtk_spin_button_init (GtkSpinButton *spin_button)
 static void
 gtk_spin_button_finalize (GObject *object)
 {
-  gtk_object_unref (GTK_OBJECT (GTK_SPIN_BUTTON (object)->adjustment));
+  gtk_spin_button_set_adjustment (GTK_SPIN_BUTTON (object), NULL);
   
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -522,7 +530,7 @@ gtk_spin_button_realize (GtkWidget *widget)
   GdkWindowAttr attributes;
   gint attributes_mask;
   guint real_width;
-  gint return_val;
+  gboolean return_val;
   gint arrow_size;
 
   spin_button = GTK_SPIN_BUTTON (widget);
@@ -562,8 +570,7 @@ gtk_spin_button_realize (GtkWidget *widget)
   gtk_style_set_background (widget->style, spin_button->panel, GTK_STATE_NORMAL);
 
   return_val = FALSE;
-  gtk_signal_emit (GTK_OBJECT (spin_button), spinbutton_signals[OUTPUT],
-		   &return_val);
+  g_signal_emit (spin_button, spinbutton_signals[OUTPUT], 0, &return_val);
   if (return_val == FALSE)
     gtk_spin_button_default_output (spin_button);
 
@@ -645,7 +652,8 @@ gtk_spin_button_size_request (GtkWidget      *widget,
 					   pango_context_get_language (context));
 
       digit_width = pango_font_metrics_get_approximate_digit_width (metrics);
-      digit_width = PANGO_PIXELS (digit_width);
+      digit_width = PANGO_SCALE *
+        ((digit_width + PANGO_SCALE - 1) / PANGO_SCALE);
 
       pango_font_metrics_unref (metrics);
       
@@ -657,15 +665,15 @@ gtk_spin_button_size_request (GtkWidget      *widget,
 
       string_len = compute_double_length (spin_button->adjustment->upper,
                                           spin_button->digits);
-      w = MIN (string_len, max_string_len) * digit_width;
+      w = PANGO_PIXELS (MIN (string_len, max_string_len) * digit_width);
       width = MAX (width, w);
       string_len = compute_double_length (spin_button->adjustment->lower,
 					  spin_button->digits);
-      w = MIN (string_len, max_string_len) * digit_width;
+      w = PANGO_PIXELS (MIN (string_len, max_string_len) * digit_width);
       width = MAX (width, w);
       
-      requisition->width = width;
-      if (interior_focus)
+      requisition->width = width + /* INNER_BORDER */ 2 * 2;
+      if (!interior_focus)
 	requisition->width += 2 * focus_width;
     }
 
@@ -750,7 +758,7 @@ gtk_spin_button_expose (GtkWidget      *widget,
       rect.x = 0;
       rect.y = 0;
 
-      gdk_window_get_size (spin->panel, &rect.width, &rect.height);
+      gdk_drawable_get_size (spin->panel, &rect.width, &rect.height);
 
       shadow_type = spin_button_get_shadow_type (spin);
       
@@ -950,16 +958,26 @@ static void
 gtk_spin_button_grab_notify (GtkWidget *widget,
 			     gboolean   was_grabbed)
 {
+  GtkSpinButton *spin = GTK_SPIN_BUTTON (widget);
+
   if (!was_grabbed)
-    gtk_spin_button_stop_spinning (GTK_SPIN_BUTTON (widget));
+    {
+      gtk_spin_button_stop_spinning (spin);
+      spin_button_redraw (spin);
+    }
 }
 
 static void
 gtk_spin_button_state_changed (GtkWidget    *widget,
 			       GtkStateType  previous_state)
 {
+  GtkSpinButton *spin = GTK_SPIN_BUTTON (widget);
+
   if (!GTK_WIDGET_IS_SENSITIVE (widget))
-    gtk_spin_button_stop_spinning (GTK_SPIN_BUTTON (widget));    
+    {
+      gtk_spin_button_stop_spinning (spin);    
+      spin_button_redraw (spin);
+    }
 }
 
 static gint
@@ -991,7 +1009,7 @@ gtk_spin_button_stop_spinning (GtkSpinButton *spin)
 {
   if (spin->timer)
     {
-      gtk_timeout_remove (spin->timer);
+      g_source_remove (spin->timer);
       spin->timer = 0;
       spin->timer_calls = 0;
       spin->need_timer = FALSE;
@@ -1001,12 +1019,15 @@ gtk_spin_button_stop_spinning (GtkSpinButton *spin)
   spin->timer = 0;
   spin->timer_step = spin->adjustment->step_increment;
   spin->timer_calls = 0;
+
+  spin->click_child = NO_ARROW;
+  spin->button = 0;
 }
 
 static void
 start_spinning (GtkSpinButton *spin,
 		GtkArrowType   click_child,
-		gfloat         step)
+		gdouble        step)
 {
   g_return_if_fail (click_child == GTK_ARROW_UP || click_child == GTK_ARROW_DOWN);
   
@@ -1017,8 +1038,9 @@ start_spinning (GtkSpinButton *spin,
     {
       spin->timer_step = step;
       spin->need_timer = TRUE;
-      spin->timer = gtk_timeout_add (SPIN_BUTTON_INITIAL_TIMER_DELAY, 
-				     (GtkFunction) gtk_spin_button_timer, (gpointer) spin);
+      spin->timer = g_timeout_add (SPIN_BUTTON_INITIAL_TIMER_DELAY, 
+				   (GSourceFunc) gtk_spin_button_timer, 
+				   (gpointer) spin);
     }
 
   spin_button_redraw (spin);
@@ -1078,7 +1100,7 @@ gtk_spin_button_button_release (GtkWidget      *widget,
 
   if (event->button == spin->button)
     {
-      guint click_child;
+      int click_child = spin->click_child;
 
       gtk_spin_button_stop_spinning (spin);
 
@@ -1088,7 +1110,7 @@ gtk_spin_button_button_release (GtkWidget      *widget,
 	      event->y <= widget->requisition.height &&
 	      event->x <= arrow_size + 2 * widget->style->xthickness)
 	    {
-	      if (spin->click_child == GTK_ARROW_UP &&
+	      if (click_child == GTK_ARROW_UP &&
 		  event->y <= widget->requisition.height / 2)
 		{
 		  gdouble diff;
@@ -1097,7 +1119,7 @@ gtk_spin_button_button_release (GtkWidget      *widget,
 		  if (diff > EPSILON)
 		    gtk_spin_button_real_spin (spin, diff);
 		}
-	      else if (spin->click_child == GTK_ARROW_DOWN &&
+	      else if (click_child == GTK_ARROW_DOWN &&
 		       event->y > widget->requisition.height / 2)
 		{
 		  gdouble diff;
@@ -1108,17 +1130,12 @@ gtk_spin_button_button_release (GtkWidget      *widget,
 		}
 	    }
 	}		  
-      click_child = spin->click_child;
-      spin->click_child = NO_ARROW;
-      spin->button = 0;
       spin_button_redraw (spin);
 
       return TRUE;
     }
   else
     return GTK_WIDGET_CLASS (parent_class)->button_release_event (widget, event);
-
-  return FALSE;
 }
 
 static gint
@@ -1172,9 +1189,9 @@ gtk_spin_button_timer (GtkSpinButton *spin_button)
       if (spin_button->need_timer)
 	{
 	  spin_button->need_timer = FALSE;
-	  spin_button->timer = gtk_timeout_add 
-	    (SPIN_BUTTON_TIMER_DELAY, (GtkFunction) gtk_spin_button_timer, 
-	     (gpointer) spin_button);
+	  spin_button->timer = g_timeout_add (SPIN_BUTTON_TIMER_DELAY, 
+					      (GSourceFunc) gtk_spin_button_timer, 
+					      (gpointer) spin_button);
 	}
       else 
 	{
@@ -1202,16 +1219,16 @@ static void
 gtk_spin_button_value_changed (GtkAdjustment *adjustment,
 			       GtkSpinButton *spin_button)
 {
-  gint return_val;
+  gboolean return_val;
+
+  g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment));
 
   return_val = FALSE;
-  gtk_signal_emit (GTK_OBJECT (spin_button), spinbutton_signals[OUTPUT],
-		   &return_val);
+  g_signal_emit (spin_button, spinbutton_signals[OUTPUT], 0, &return_val);
   if (return_val == FALSE)
     gtk_spin_button_default_output (spin_button);
 
-  gtk_signal_emit (GTK_OBJECT (spin_button), 
-		   spinbutton_signals[VALUE_CHANGED]);
+  g_signal_emit (spin_button, spinbutton_signals[VALUE_CHANGED], 0);
 
   spin_button_redraw (spin_button);
   
@@ -1222,81 +1239,83 @@ static void
 gtk_spin_button_real_change_value (GtkSpinButton *spin,
 				   GtkScrollType  scroll)
 {
-  if (GTK_ENTRY (spin)->editable)
+  /* We don't test whether the entry is editable, since
+   * this key binding conceptually corresponds to changing
+   * the value with the buttons using the mouse, which
+   * we allow for non-editable spin buttons.
+   */
+  switch (scroll)
     {
-      switch (scroll)
-        {
-        case GTK_SCROLL_STEP_BACKWARD:
-        case GTK_SCROLL_STEP_DOWN:
-        case GTK_SCROLL_STEP_LEFT:
-	  gtk_spin_button_real_spin (spin, -spin->timer_step);
-	  
-	  if (spin->climb_rate > 0.0 && spin->timer_step
-	      < spin->adjustment->page_increment)
-	    {
-	      if (spin->timer_calls < MAX_TIMER_CALLS)
-		spin->timer_calls++;
-	      else 
-		{
-		  spin->timer_calls = 0;
-		  spin->timer_step += spin->climb_rate;
-		}
-	    }
-	  break;
-
-        case GTK_SCROLL_STEP_FORWARD:
-        case GTK_SCROLL_STEP_UP:
-        case GTK_SCROLL_STEP_RIGHT:
-	  gtk_spin_button_real_spin (spin, spin->timer_step);
-	  
-	  if (spin->climb_rate > 0.0 && spin->timer_step
-	      < spin->adjustment->page_increment)
-	    {
-	      if (spin->timer_calls < MAX_TIMER_CALLS)
-		spin->timer_calls++;
-	      else 
-		{
-		  spin->timer_calls = 0;
-		  spin->timer_step += spin->climb_rate;
-		}
-	    }
-          break;
-
-        case GTK_SCROLL_PAGE_BACKWARD:
-        case GTK_SCROLL_PAGE_DOWN:
-        case GTK_SCROLL_PAGE_LEFT:
-	  gtk_spin_button_real_spin (spin, -spin->adjustment->page_increment);
-          break;
-
-        case GTK_SCROLL_PAGE_FORWARD:
-        case GTK_SCROLL_PAGE_UP:
-        case GTK_SCROLL_PAGE_RIGHT:
-	  gtk_spin_button_real_spin (spin, spin->adjustment->page_increment);
-          break;
-          
-	case GTK_SCROLL_START:
-	  {
-	    gdouble diff = spin->adjustment->value - spin->adjustment->lower;
-	    if (diff > EPSILON)
-	      gtk_spin_button_real_spin (spin, -diff);
-	    break;
-	  }
-
-	case GTK_SCROLL_END:
-	  {
-	    gdouble diff = spin->adjustment->upper - spin->adjustment->value;
-	    if (diff > EPSILON)
-	      gtk_spin_button_real_spin (spin, diff);
-	    break;
-	  }
-	  
-        default:
-	  g_warning ("Invalid scroll type %d for GtkSpinButton::change-value", scroll);
-          break;
-        }
+    case GTK_SCROLL_STEP_BACKWARD:
+    case GTK_SCROLL_STEP_DOWN:
+    case GTK_SCROLL_STEP_LEFT:
+      gtk_spin_button_real_spin (spin, -spin->timer_step);
       
-      gtk_spin_button_update (spin);
+      if (spin->climb_rate > 0.0 && spin->timer_step
+	  < spin->adjustment->page_increment)
+	{
+	  if (spin->timer_calls < MAX_TIMER_CALLS)
+	    spin->timer_calls++;
+	  else 
+	    {
+	      spin->timer_calls = 0;
+	      spin->timer_step += spin->climb_rate;
+	    }
+	}
+      break;
+      
+    case GTK_SCROLL_STEP_FORWARD:
+    case GTK_SCROLL_STEP_UP:
+    case GTK_SCROLL_STEP_RIGHT:
+      gtk_spin_button_real_spin (spin, spin->timer_step);
+      
+      if (spin->climb_rate > 0.0 && spin->timer_step
+	  < spin->adjustment->page_increment)
+	{
+	  if (spin->timer_calls < MAX_TIMER_CALLS)
+	    spin->timer_calls++;
+	  else 
+	    {
+	      spin->timer_calls = 0;
+	      spin->timer_step += spin->climb_rate;
+	    }
+	}
+      break;
+      
+    case GTK_SCROLL_PAGE_BACKWARD:
+    case GTK_SCROLL_PAGE_DOWN:
+    case GTK_SCROLL_PAGE_LEFT:
+      gtk_spin_button_real_spin (spin, -spin->adjustment->page_increment);
+      break;
+      
+    case GTK_SCROLL_PAGE_FORWARD:
+    case GTK_SCROLL_PAGE_UP:
+    case GTK_SCROLL_PAGE_RIGHT:
+      gtk_spin_button_real_spin (spin, spin->adjustment->page_increment);
+      break;
+      
+    case GTK_SCROLL_START:
+      {
+	gdouble diff = spin->adjustment->value - spin->adjustment->lower;
+	if (diff > EPSILON)
+	  gtk_spin_button_real_spin (spin, -diff);
+	break;
+      }
+      
+    case GTK_SCROLL_END:
+      {
+	gdouble diff = spin->adjustment->upper - spin->adjustment->value;
+	if (diff > EPSILON)
+	  gtk_spin_button_real_spin (spin, diff);
+	break;
+      }
+      
+    default:
+      g_warning ("Invalid scroll type %d for GtkSpinButton::change-value", scroll);
+      break;
     }
+  
+  gtk_spin_button_update (spin);
 }
 
 static gint
@@ -1320,6 +1339,9 @@ gtk_spin_button_snap (GtkSpinButton *spin_button,
   gdouble tmp;
 
   inc = spin_button->adjustment->step_increment;
+  if (inc == 0)
+    return;
+  
   tmp = (val - spin_button->adjustment->lower) / inc;
   if (tmp - floor (tmp) < ceil (tmp) - tmp)
     val = spin_button->adjustment->lower + floor (tmp) * inc;
@@ -1331,8 +1353,7 @@ gtk_spin_button_snap (GtkSpinButton *spin_button,
   else
     {
       gint return_val = FALSE;
-      gtk_signal_emit (GTK_OBJECT (spin_button), spinbutton_signals[OUTPUT],
-		       &return_val);
+      g_signal_emit (spin_button, spinbutton_signals[OUTPUT], 0, &return_val);
       if (return_val == FALSE)
 	gtk_spin_button_default_output (spin_button);
     }
@@ -1343,6 +1364,9 @@ gtk_spin_button_activate (GtkEntry *entry)
 {
   if (entry->editable)
     gtk_spin_button_update (GTK_SPIN_BUTTON (entry));
+
+  /* Chain up so that entry->activates_default is honored */
+  parent_class->activate (entry);
 }
 
 static void
@@ -1475,7 +1499,7 @@ gtk_spin_button_default_input (GtkSpinButton *spin_button,
 {
   gchar *err = NULL;
 
-  *new_val = strtod (gtk_entry_get_text (GTK_ENTRY (spin_button)), &err);
+  *new_val = g_strtod (gtk_entry_get_text (GTK_ENTRY (spin_button)), &err);
   if (*err)
     return GTK_INPUT_ERROR;
   else
@@ -1541,7 +1565,7 @@ gtk_spin_button_new (GtkAdjustment *adjustment,
   if (adjustment)
     g_return_val_if_fail (GTK_IS_ADJUSTMENT (adjustment), NULL);
 
-  spin = gtk_type_new (GTK_TYPE_SPIN_BUTTON);
+  spin = g_object_new (GTK_TYPE_SPIN_BUTTON, NULL);
 
   gtk_spin_button_configure (spin, adjustment, climb_rate, digits);
 
@@ -1574,7 +1598,7 @@ gtk_spin_button_new_with_range (gdouble min,
   g_return_val_if_fail (min < max, NULL);
   g_return_val_if_fail (step != 0.0, NULL);
 
-  spin = gtk_type_new (GTK_TYPE_SPIN_BUTTON);
+  spin = g_object_new (GTK_TYPE_SPIN_BUTTON, NULL);
 
   adj = gtk_adjustment_new (min, min, max, step, 10 * step, 0);
 
@@ -1624,26 +1648,29 @@ gtk_spin_button_set_adjustment (GtkSpinButton *spin_button,
     {
       if (spin_button->adjustment)
         {
-          gtk_signal_disconnect_by_data (GTK_OBJECT (spin_button->adjustment),
-                                         (gpointer) spin_button);
-          gtk_object_unref (GTK_OBJECT (spin_button->adjustment));
+	  g_signal_handlers_disconnect_by_func (spin_button->adjustment,
+						gtk_spin_button_value_changed,
+						spin_button);
+	  g_signal_handlers_disconnect_by_func (spin_button->adjustment,
+						adjustment_changed_cb,
+						spin_button);
+	  g_object_unref (spin_button->adjustment);
         }
       spin_button->adjustment = adjustment;
       if (adjustment)
         {
-          gtk_object_ref (GTK_OBJECT (adjustment));
+	  g_object_ref (adjustment);
 	  gtk_object_sink (GTK_OBJECT (adjustment));
-          gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
-			      (GtkSignalFunc) gtk_spin_button_value_changed,
-			      (gpointer) spin_button);
-	  gtk_signal_connect (GTK_OBJECT (adjustment), "changed",
-			      (GtkSignalFunc) adjustment_changed_cb,
-			      (gpointer) spin_button);
+	  g_signal_connect (adjustment, "value_changed",
+			    G_CALLBACK (gtk_spin_button_value_changed),
+			    spin_button);
+	  g_signal_connect (adjustment, "changed",
+			    G_CALLBACK (adjustment_changed_cb),
+			    spin_button);
+	  spin_button->timer_step = spin_button->adjustment->step_increment;
         }
 
       gtk_widget_queue_resize (GTK_WIDGET (spin_button));
-
-      spin_button->timer_step = spin_button->adjustment->step_increment;
     }
 
   g_object_notify (G_OBJECT (spin_button), "adjustment");
@@ -1668,7 +1695,7 @@ gtk_spin_button_get_adjustment (GtkSpinButton *spin_button)
 /**
  * gtk_spin_button_set_digits:
  * @spin_button: a #GtkSpinButton
- * @digits: the number of digits to be displayed for the spin button's value
+ * @digits: the number of digits after the decimal point to be displayed for the spin button's value
  * 
  * Set the precision to be displayed by @spin_button. Up to 20 digit precision
  * is allowed.
@@ -1856,8 +1883,7 @@ gtk_spin_button_set_value (GtkSpinButton *spin_button,
   else
     {
       gint return_val = FALSE;
-      gtk_signal_emit (GTK_OBJECT (spin_button), spinbutton_signals[OUTPUT],
-		       &return_val);
+      g_signal_emit (spin_button, spinbutton_signals[OUTPUT], 0, &return_val);
       if (return_val == FALSE)
 	gtk_spin_button_default_output (spin_button);
     }
@@ -2152,8 +2178,7 @@ gtk_spin_button_update (GtkSpinButton *spin_button)
   g_return_if_fail (GTK_IS_SPIN_BUTTON (spin_button));
 
   return_val = FALSE;
-  gtk_signal_emit (GTK_OBJECT (spin_button), spinbutton_signals[INPUT],
-		   &val, &return_val);
+  g_signal_emit (spin_button, spinbutton_signals[INPUT], 0, &val, &return_val);
   if (return_val == FALSE)
     {
       return_val = gtk_spin_button_default_input (spin_button, &val);
@@ -2189,8 +2214,8 @@ gtk_spin_button_update (GtkSpinButton *spin_button)
       else
 	{
 	  return_val = FALSE;
-	  gtk_signal_emit (GTK_OBJECT (spin_button), spinbutton_signals[OUTPUT],
-			   &return_val);
+	  g_signal_emit (spin_button, spinbutton_signals[OUTPUT], 0,
+			 &return_val);
 	  if (return_val == FALSE)
 	    gtk_spin_button_default_output (spin_button);
 	}
