@@ -92,6 +92,7 @@ struct _GtkTextRenderState
   GtkTextAppearance *last_bg_appearance;
   GdkGC *fg_gc;
   GdkGC *bg_gc;
+  GdkGC *error_gc;
   GdkRectangle clip_rect;
 };
 
@@ -116,6 +117,7 @@ gtk_text_render_state_new (GtkWidget    *widget,
   state->widget = widget;
   state->fg_gc = gdk_gc_new (drawable);
   state->bg_gc = gdk_gc_new (drawable);
+
   state->clip_rect = *clip_rect;
 
   return state;
@@ -126,6 +128,8 @@ gtk_text_render_state_destroy (GtkTextRenderState *state)
 {
   g_object_unref (state->fg_gc);
   g_object_unref (state->bg_gc);
+  if (state->error_gc)
+    g_object_unref (state->error_gc);
 
   g_free (state);
 }
@@ -204,6 +208,36 @@ gtk_text_render_state_update (GtkTextRenderState *state,
   state->last_appearance = new_appearance;
 }
 
+static GdkGC *
+gtk_text_render_state_get_error_gc (GtkTextRenderState *state)
+{
+  if (!state->error_gc)
+    {
+      static const GdkColor red = { 0, 0xffff, 0, 0 };
+      
+      GdkGCValues gc_values;
+      GdkGCValuesMask gc_values_mask;
+      GdkColor *underline_color;
+      GtkWidget *widget = state->widget;
+
+      gtk_widget_style_get (widget, "error-underline_color", &underline_color, NULL);
+      
+      gc_values_mask = GDK_GC_FOREGROUND;
+      if (underline_color)
+	{
+	  gc_values.foreground = *underline_color;
+	  gdk_color_free (underline_color);
+	}
+      else
+	gc_values.foreground = red;
+      
+      gdk_rgb_find_color (widget->style->colormap, &gc_values.foreground);
+      state->error_gc = gdk_gc_new_with_values (widget->window, &gc_values, gc_values_mask);
+    }
+
+  return state->error_gc;
+}
+
 static void
 get_shape_extents (PangoLayoutRun *run,
                    PangoRectangle *ink_rect,
@@ -423,6 +457,29 @@ render_layout_line (GdkDrawable        *drawable,
                          x + (x_off + ink_rect.x + ink_rect.width) / PANGO_SCALE,
                          risen_y + 1);
           break;
+        case PANGO_UNDERLINE_ERROR:
+          g_assert (need_ink);
+          {
+	    GdkGC *error_gc = gtk_text_render_state_get_error_gc (render_state);
+
+            int point_x, point_y;
+            int counter = 0;
+	    int end_x = x + (x_off + ink_rect.x + ink_rect.width) / PANGO_SCALE;
+
+            for (point_x = x + (x_off + ink_rect.x) / PANGO_SCALE - 1;
+                 point_x <= end_x;
+                 point_x += 2)
+	      {
+		point_y = counter ? risen_y + 1 : risen_y + 2;
+		
+		gdk_draw_line (drawable, error_gc,
+			       point_x, point_y,
+			       MIN (point_x + 1, end_x), point_y);
+		
+		counter = (counter + 1) % 2;
+	      }
+          }
+          break;
         case PANGO_UNDERLINE_LOW:
           g_assert (need_ink);
           gdk_draw_line (drawable, fg_gc,
@@ -821,7 +878,6 @@ gtk_text_layout_draw (GtkTextLayout *layout,
               GtkTextCursorDisplay *cursor = cursor_list->data;
 	      GtkTextDirection dir;
  	      GdkRectangle cursor_location;
-              GdkGC *gc;
 
               dir = line_display->direction;
  	      if (have_strong && have_weak)
@@ -834,14 +890,10 @@ gtk_text_layout_draw (GtkTextLayout *layout,
  	      cursor_location.y = current_y + line_display->top_margin + cursor->y;
  	      cursor_location.width = 0;
  	      cursor_location.height = cursor->height;
- 
-	      gc = _gtk_get_insertion_cursor_gc (widget, cursor->is_strong);
-	      gdk_gc_set_clip_rectangle(gc, &clip);
- 	      _gtk_draw_insertion_cursor (widget, drawable, gc, &cursor_location,
-                                          dir, have_strong && have_weak);
-              gdk_gc_set_clip_rectangle (gc, NULL);
 
-	      g_object_unref (gc);
+	      gtk_draw_insertion_cursor (widget, drawable, &clip, &cursor_location,
+					 cursor->is_strong,
+					 dir, have_strong && have_weak);
 
               cursor_list = cursor_list->next;
             }
