@@ -24,6 +24,8 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#undef GTK_DISABLE_DEPRECATED
+
 #include "gtklabel.h"
 #include "gtkeventbox.h"
 #include "gtkpixmap.h"
@@ -382,7 +384,7 @@ gtk_tree_item_add_pixmaps (GtkTreeItem *tree_item)
       pixmap_node = g_new (GtkTreePixmaps, 1);
 
       pixmap_node->colormap = colormap;
-      gdk_colormap_ref (colormap);
+      g_object_ref (colormap);
 
       pixmap_node->refcount = 1;
 
@@ -391,14 +393,14 @@ gtk_tree_item_add_pixmaps (GtkTreeItem *tree_item)
 	gdk_pixmap_create_from_xpm_d (GTK_WIDGET (tree_item)->window,
 				      &pixmap_node->mask_plus,
 				      NULL,
-				      tree_plus);
+				      (gchar **)tree_plus);
       
       /* create pixmaps for minus icon */
       pixmap_node->pixmap_minus = 
 	gdk_pixmap_create_from_xpm_d (GTK_WIDGET (tree_item)->window,
 				      &pixmap_node->mask_minus,
 				      NULL,
-				      tree_minus);
+				      (gchar **)tree_minus);
 
       tree_item->pixmaps = pixmaps = g_list_prepend (pixmaps, pixmap_node);
     }
@@ -422,11 +424,11 @@ gtk_tree_item_remove_pixmaps (GtkTreeItem *tree_item)
       
       if (--pixmap_node->refcount == 0)
 	{
-	  gdk_colormap_unref (pixmap_node->colormap);
-	  gdk_pixmap_unref (pixmap_node->pixmap_plus);
-	  gdk_bitmap_unref (pixmap_node->mask_plus);
-	  gdk_pixmap_unref (pixmap_node->pixmap_minus);
-	  gdk_bitmap_unref (pixmap_node->mask_minus);
+	  g_object_unref (pixmap_node->colormap);
+	  g_object_unref (pixmap_node->pixmap_plus);
+	  g_object_unref (pixmap_node->mask_plus);
+	  g_object_unref (pixmap_node->pixmap_minus);
+	  g_object_unref (pixmap_node->mask_minus);
 	  
 	  pixmaps = g_list_remove_link (pixmaps, tree_item->pixmaps);
 	  g_list_free_1 (tree_item->pixmaps);
@@ -634,7 +636,7 @@ gtk_tree_item_paint (GtkWidget    *widget,
 	{
 	  if (!GTK_WIDGET_IS_SENSITIVE (widget)) 
 	    gtk_paint_flat_box(widget->style, widget->window,
-			       widget->state, GTK_STATE_INSENSITIVE,
+			       widget->state, GTK_SHADOW_NONE,
 			       area, widget, "treeitem",
 			       0, 0, -1, -1);
 	  else
@@ -664,11 +666,11 @@ gtk_tree_item_paint (GtkWidget    *widget,
 	}
 
       if (GTK_WIDGET_HAS_FOCUS (widget))
-	gtk_paint_focus (widget->style, widget->window,
+	gtk_paint_focus (widget->style, widget->window, GTK_WIDGET_STATE (widget),
 			 NULL, widget, "treeitem",
 			 0, 0,
-			 widget->allocation.width - 1,
-			 widget->allocation.height - 1);
+			 widget->allocation.width,
+			 widget->allocation.height);
       
     }
 }
@@ -689,10 +691,43 @@ gtk_tree_item_button_press (GtkWidget      *widget,
   return (event->type == GDK_BUTTON_PRESS && GTK_WIDGET_IS_SENSITIVE(widget));
 }
 
+static void
+gtk_tree_item_expose_child (GtkWidget *child,
+                            gpointer   client_data)
+{
+  struct {
+    GtkWidget *container;
+    GdkEventExpose *event;
+  } *data = client_data;
+
+  if (GTK_WIDGET_DRAWABLE (child) &&
+      GTK_WIDGET_NO_WINDOW (child) &&
+      (child->window == data->event->window))
+    {
+      GdkEvent *child_event = gdk_event_new (GDK_EXPOSE);
+      child_event->expose = *data->event;
+      g_object_ref (child_event->expose.window);
+
+      child_event->expose.region = gtk_widget_region_intersect (child,
+								data->event->region);
+      if (!gdk_region_empty (child_event->expose.region))
+        {
+          gdk_region_get_clipbox (child_event->expose.region, &child_event->expose.area);
+          gtk_widget_send_expose (child, child_event);
+	}
+      gdk_event_free (child_event);
+    }
+}
+
 static gint
 gtk_tree_item_expose (GtkWidget      *widget,
 		      GdkEventExpose *event)
 {
+  struct {
+    GtkWidget *container;
+    GdkEventExpose *event;
+  } data;
+  
   g_return_val_if_fail (GTK_IS_TREE_ITEM (widget), FALSE);
   g_return_val_if_fail (event != NULL, FALSE);
 
@@ -700,7 +735,12 @@ gtk_tree_item_expose (GtkWidget      *widget,
     {
       gtk_tree_item_paint (widget, &event->area);
 
-      (* GTK_WIDGET_CLASS (parent_class)->expose_event) (widget, event);
+      data.container = widget;
+      data.event = event;
+
+      gtk_container_forall (GTK_CONTAINER (widget),
+                            gtk_tree_item_expose_child,
+                            &data);
    }
 
   return FALSE;
