@@ -118,7 +118,7 @@ gtk_tree_selection_finalize (GObject *object)
  * _gtk_tree_selection_new:
  *
  * Creates a new #GtkTreeSelection object.  This function should not be invoked,
- * as each #GtkTreeView will create it's own #GtkTreeSelection.
+ * as each #GtkTreeView will create its own #GtkTreeSelection.
  *
  * Return value: A newly created #GtkTreeSelection object.
  **/
@@ -137,7 +137,7 @@ _gtk_tree_selection_new (void)
  * @tree_view: The #GtkTreeView.
  *
  * Creates a new #GtkTreeSelection object.  This function should not be invoked,
- * as each #GtkTreeView will create it's own #GtkTreeSelection.
+ * as each #GtkTreeView will create its own #GtkTreeSelection.
  *
  * Return value: A newly created #GtkTreeSelection object.
  **/
@@ -241,7 +241,7 @@ gtk_tree_selection_set_mode (GtkTreeSelection *selection,
 						  node,
 						  tree,
 						  anchor_path,
-						  0,
+                                                  0,
 						  FALSE);
       if (anchor_path)
 	gtk_tree_path_free (anchor_path);
@@ -412,7 +412,7 @@ gtk_tree_selection_get_selected (GtkTreeSelection  *selection,
  *
  * Creates a list of path of all selected rows. Additionally, if you are
  * planning on modifying the model after calling this function, you may
- * want to convert the returned list into a list of #GtkTreeRowReferences.
+ * want to convert the returned list into a list of #GtkTreeRowReference<!-- -->s.
  * To do this, you can use gtk_tree_row_reference_new_proxy().
  *
  * To free the return value, use:
@@ -422,6 +422,8 @@ gtk_tree_selection_get_selected (GtkTreeSelection  *selection,
  * </programlisting></informalexample>
  *
  * Return value: A #GList containing a #GtkTreePath for each selected row.
+ *
+ * Since: 2.2
  **/
 GList *
 gtk_tree_selection_get_selected_rows (GtkTreeSelection   *selection,
@@ -453,7 +455,7 @@ gtk_tree_selection_get_selected_rows (GtkTreeSelection   *selection,
         {
 	  GtkTreePath *path;
 
-	  path = gtk_tree_model_get_path (*model, &iter);
+	  path = gtk_tree_model_get_path (selection->tree_view->priv->model, &iter);
 	  list = g_list_append (list, path);
 
 	  return list;
@@ -472,7 +474,7 @@ gtk_tree_selection_get_selected_rows (GtkTreeSelection   *selection,
   do
     {
       if (GTK_RBNODE_FLAG_SET (node, GTK_RBNODE_IS_SELECTED))
-	list = g_list_append (list, gtk_tree_path_copy (path));
+	list = g_list_prepend (list, gtk_tree_path_copy (path));
 
       if (node->children)
         {
@@ -504,7 +506,8 @@ gtk_tree_selection_get_selected_rows (GtkTreeSelection   *selection,
 		  if (!tree)
 		    {
 		      gtk_tree_path_free (path);
-		      return list;
+
+		      goto done; 
 		    }
 
 		  gtk_tree_path_up (path);
@@ -517,7 +520,8 @@ gtk_tree_selection_get_selected_rows (GtkTreeSelection   *selection,
 
   gtk_tree_path_free (path);
 
-  return list;
+ done:
+  return g_list_reverse (list);
 }
 
 static void
@@ -543,6 +547,8 @@ gtk_tree_selection_count_selected_rows_helper (GtkRBTree *tree,
  * Returns the number of rows that have been selected in @tree.
  *
  * Return value: The number of rows selected.
+ * 
+ * Since: 2.2
  **/
 gint
 gtk_tree_selection_count_selected_rows (GtkTreeSelection *selection)
@@ -580,6 +586,15 @@ gtk_tree_selection_count_selected_rows (GtkTreeSelection *selection)
   return count;
 }
 
+/* gtk_tree_selection_selected_foreach helper */
+static void
+model_changed (gpointer data)
+{
+  gboolean *stop = (gboolean *)data;
+
+  *stop = TRUE;
+}
+
 /**
  * gtk_tree_selection_selected_foreach:
  * @selection: A #GtkTreeSelection.
@@ -600,6 +615,9 @@ gtk_tree_selection_selected_foreach (GtkTreeSelection            *selection,
   GtkRBNode *node;
   GtkTreeIter iter;
 
+  guint inserted_id, deleted_id, reordered_id;
+  gboolean stop = FALSE, has_next = TRUE, has_parent = TRUE;
+
   g_return_if_fail (GTK_IS_TREE_SELECTION (selection));
   g_return_if_fail (selection->tree_view != NULL);
   g_return_if_fail (selection->tree_view->priv->model != NULL);
@@ -628,6 +646,20 @@ gtk_tree_selection_selected_foreach (GtkTreeSelection            *selection,
   while (node->left != tree->nil)
     node = node->left;
 
+  /* connect to signals to monitor changes in treemodel */
+  inserted_id = g_signal_connect_swapped (selection->tree_view->priv->model,
+                                          "row_inserted",
+					  G_CALLBACK (model_changed),
+				          &stop);
+  deleted_id = g_signal_connect_swapped (selection->tree_view->priv->model,
+                                         "row_deleted",
+					 G_CALLBACK (model_changed),
+				         &stop);
+  reordered_id = g_signal_connect_swapped (selection->tree_view->priv->model,
+                                           "rows_reordered",
+					   G_CALLBACK (model_changed),
+				           &stop);
+
   /* find the node internally */
   path = gtk_tree_path_new_first ();
   gtk_tree_model_get_iter (selection->tree_view->priv->model,
@@ -637,6 +669,10 @@ gtk_tree_selection_selected_foreach (GtkTreeSelection            *selection,
     {
       if (GTK_RBNODE_FLAG_SET (node, GTK_RBNODE_IS_SELECTED))
 	(* func) (selection->tree_view->priv->model, path, &iter, data);
+
+      if (stop)
+	goto out;
+
       if (node->children)
 	{
 	  gboolean has_child;
@@ -649,8 +685,10 @@ gtk_tree_selection_selected_foreach (GtkTreeSelection            *selection,
 	  tmp = iter;
 	  has_child = gtk_tree_model_iter_children (selection->tree_view->priv->model, &iter, &tmp);
 	  gtk_tree_path_append_index (path, 0);
-	  /* Sanity Check! */
-	  TREE_VIEW_INTERNAL_ASSERT_VOID (has_child);
+
+	  /* we do the sanity check at the bottom of this function */
+	  if (!has_child)
+	    goto out;
 	}
       else
 	{
@@ -666,8 +704,9 @@ gtk_tree_selection_selected_foreach (GtkTreeSelection            *selection,
 		  done = TRUE;
 		  gtk_tree_path_next (path);
 
-		  /* Sanity Check! */
-		  TREE_VIEW_INTERNAL_ASSERT_VOID (has_next);
+		  /* we do the sanity check at the bottom of this function */
+		  if (!has_next)
+		    goto out;
 		}
 	      else
 		{
@@ -678,21 +717,47 @@ gtk_tree_selection_selected_foreach (GtkTreeSelection            *selection,
 		  tree = tree->parent_tree;
 		  if (tree == NULL)
 		    {
-		      gtk_tree_path_free (path);
 		      /* we've run out of tree */
 		      /* We're done with this function */
-		      return;
+
+		      goto out;
 		    }
+
 		  has_parent = gtk_tree_model_iter_parent (selection->tree_view->priv->model, &iter, &tmp_iter);
 		  gtk_tree_path_up (path);
-		  /* Sanity check */
-		  TREE_VIEW_INTERNAL_ASSERT_VOID (has_parent);
+
+		  /* we do the sanity check at the bottom of this function */
+		  if (!has_parent)
+		    goto out;
 		}
 	    }
 	  while (!done);
 	}
     }
   while (TRUE);
+
+out:
+  if (path)
+    gtk_tree_path_free (path);
+
+  g_signal_handler_disconnect (selection->tree_view->priv->model,
+                               inserted_id);
+  g_signal_handler_disconnect (selection->tree_view->priv->model,
+                               deleted_id);
+  g_signal_handler_disconnect (selection->tree_view->priv->model,
+                               reordered_id);
+
+  /* check if we have to spew a scary message */
+  if (!has_next)
+    TREE_VIEW_INTERNAL_ASSERT_VOID (has_next);
+  if (!has_parent)
+    TREE_VIEW_INTERNAL_ASSERT_VOID (has_parent);
+  if (stop)
+    g_warning
+      ("The model has been modified from within gtk_tree_selection_selected_foreach.\n"
+       "This function is for observing the selections of the tree only.  If\n"
+       "you are trying to get all selected items from the tree, try using\n"
+       "gtk_tree_selection_get_selected_rows instead.\n");
 }
 
 /**
@@ -708,8 +773,8 @@ gtk_tree_selection_select_path (GtkTreeSelection *selection,
 {
   GtkRBNode *node;
   GtkRBTree *tree;
-  GdkModifierType state = 0;
   gboolean ret;
+  GtkTreeSelectMode mode = 0;
 
   g_return_if_fail (GTK_IS_TREE_SELECTION (selection));
   g_return_if_fail (selection->tree_view != NULL);
@@ -725,13 +790,13 @@ gtk_tree_selection_select_path (GtkTreeSelection *selection,
     return;
 
   if (selection->type == GTK_SELECTION_MULTIPLE)
-    state = GDK_CONTROL_MASK;
+    mode = GTK_TREE_SELECT_MODE_TOGGLE;
 
   _gtk_tree_selection_internal_select_node (selection,
 					    node,
 					    tree,
 					    path,
-					    state,
+                                            mode,
 					    FALSE);
 }
 
@@ -767,7 +832,7 @@ gtk_tree_selection_unselect_path (GtkTreeSelection *selection,
 					    node,
 					    tree,
 					    path,
-					    GDK_CONTROL_MASK,
+                                            GTK_TREE_SELECT_MODE_TOGGLE,
 					    TRUE);
 }
 
@@ -1088,6 +1153,7 @@ gtk_tree_selection_real_modify_range (GtkTreeSelection *selection,
 {
   GtkRBNode *start_node, *end_node;
   GtkRBTree *start_tree, *end_tree;
+  GtkTreePath *anchor_path = NULL;
   gboolean dirty = FALSE;
 
   switch (gtk_tree_path_compare (start_path, end_path))
@@ -1101,6 +1167,7 @@ gtk_tree_selection_real_modify_range (GtkTreeSelection *selection,
 				start_path,
 				&end_tree,
 				&end_node);
+      anchor_path = start_path;
       break;
     case 0:
       _gtk_tree_view_find_node (selection->tree_view,
@@ -1109,6 +1176,7 @@ gtk_tree_selection_real_modify_range (GtkTreeSelection *selection,
 				&start_node);
       end_tree = start_tree;
       end_node = start_node;
+      anchor_path = start_path;
       break;
     case -1:
       _gtk_tree_view_find_node (selection->tree_view,
@@ -1119,12 +1187,24 @@ gtk_tree_selection_real_modify_range (GtkTreeSelection *selection,
 				end_path,
 				&end_tree,
 				&end_node);
+      anchor_path = start_path;
       break;
     }
 
   g_return_val_if_fail (start_node != NULL, FALSE);
   g_return_val_if_fail (end_node != NULL, FALSE);
 
+  if (anchor_path)
+    {
+      if (selection->tree_view->priv->anchor)
+	gtk_tree_row_reference_free (selection->tree_view->priv->anchor);
+
+      selection->tree_view->priv->anchor =
+	gtk_tree_row_reference_new_proxy (G_OBJECT (selection->tree_view),
+	                                  selection->tree_view->priv->model,
+					  anchor_path);
+    }
+
   do
     {
       dirty |= gtk_tree_selection_real_select_node (selection, start_tree, start_node, (mode == RANGE_SELECT)?TRUE:FALSE);
@@ -1183,6 +1263,8 @@ gtk_tree_selection_select_range (GtkTreeSelection *selection,
  *
  * Unselects a range of nodes, determined by @start_path and @end_path
  * inclusive.
+ *
+ * Since: 2.2
  **/
 void
 gtk_tree_selection_unselect_range (GtkTreeSelection *selection,
@@ -1209,7 +1291,7 @@ _gtk_tree_selection_internal_select_node (GtkTreeSelection *selection,
 					  GtkRBNode        *node,
 					  GtkRBTree        *tree,
 					  GtkTreePath      *path,
-					  GdkModifierType   state,
+                                          GtkTreeSelectMode mode,
 					  gboolean          override_browse_mode)
 {
   gint flags;
@@ -1234,7 +1316,7 @@ _gtk_tree_selection_internal_select_node (GtkTreeSelection *selection,
       else if (selection->type == GTK_SELECTION_SINGLE &&
 	       anchor_path && gtk_tree_path_compare (path, anchor_path) == 0)
 	{
-	  if ((state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK)
+	  if ((mode & GTK_TREE_SELECT_MODE_TOGGLE) == GTK_TREE_SELECT_MODE_TOGGLE)
 	    {
 	      dirty = gtk_tree_selection_real_unselect_all (selection);
 	    }
@@ -1268,7 +1350,11 @@ _gtk_tree_selection_internal_select_node (GtkTreeSelection *selection,
 	      if (dirty)
 		{
 		  if (selection->tree_view->priv->anchor)
-		    gtk_tree_row_reference_free (selection->tree_view->priv->anchor);
+                    {
+                      gtk_tree_row_reference_free (selection->tree_view->priv->anchor);
+                      selection->tree_view->priv->anchor = NULL;
+                    }
+
 		  if (gtk_tree_selection_real_select_node (selection, tree, node, TRUE))
 		    {
 		      selection->tree_view->priv->anchor =
@@ -1281,6 +1367,9 @@ _gtk_tree_selection_internal_select_node (GtkTreeSelection *selection,
 	      if (gtk_tree_selection_real_select_node (selection, tree, node, TRUE))
 		{
 		  dirty = TRUE;
+		  if (selection->tree_view->priv->anchor)
+		    gtk_tree_row_reference_free (selection->tree_view->priv->anchor);
+
 		  selection->tree_view->priv->anchor =
 		    gtk_tree_row_reference_new_proxy (G_OBJECT (selection->tree_view), selection->tree_view->priv->model, path);
 		}
@@ -1289,7 +1378,8 @@ _gtk_tree_selection_internal_select_node (GtkTreeSelection *selection,
     }
   else if (selection->type == GTK_SELECTION_MULTIPLE)
     {
-      if (((state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK) && (anchor_path == NULL))
+      if ((mode & GTK_TREE_SELECT_MODE_EXTEND) == GTK_TREE_SELECT_MODE_EXTEND
+          && (anchor_path == NULL))
 	{
 	  if (selection->tree_view->priv->anchor)
 	    gtk_tree_row_reference_free (selection->tree_view->priv->anchor);
@@ -1298,13 +1388,13 @@ _gtk_tree_selection_internal_select_node (GtkTreeSelection *selection,
 	    gtk_tree_row_reference_new_proxy (G_OBJECT (selection->tree_view), selection->tree_view->priv->model, path);
 	  dirty = gtk_tree_selection_real_select_node (selection, tree, node, TRUE);
 	}
-      else if ((state & (GDK_CONTROL_MASK|GDK_SHIFT_MASK)) == (GDK_SHIFT_MASK|GDK_CONTROL_MASK))
+      else if ((mode & (GTK_TREE_SELECT_MODE_EXTEND | GTK_TREE_SELECT_MODE_TOGGLE)) == (GTK_TREE_SELECT_MODE_EXTEND | GTK_TREE_SELECT_MODE_TOGGLE))
 	{
 	  gtk_tree_selection_select_range (selection,
 					   anchor_path,
 					   path);
 	}
-      else if ((state & GDK_CONTROL_MASK) == GDK_CONTROL_MASK)
+      else if ((mode & GTK_TREE_SELECT_MODE_TOGGLE) == GTK_TREE_SELECT_MODE_TOGGLE)
 	{
 	  flags = node->flags;
 	  if (selection->tree_view->priv->anchor)
@@ -1318,7 +1408,7 @@ _gtk_tree_selection_internal_select_node (GtkTreeSelection *selection,
 	  else
 	    dirty |= gtk_tree_selection_real_select_node (selection, tree, node, TRUE);
 	}
-      else if ((state & GDK_SHIFT_MASK) == GDK_SHIFT_MASK)
+      else if ((mode & GTK_TREE_SELECT_MODE_EXTEND) == GTK_TREE_SELECT_MODE_EXTEND)
 	{
 	  dirty = gtk_tree_selection_real_unselect_all (selection);
 	  dirty |= gtk_tree_selection_real_modify_range (selection,
